#!/usr/bin/env python

"""
View annotated sources.

Copyright (C) 2006, 2007, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.data import *
from compiler.visitor import ASTVisitor
import sys
import os
import textwrap

# Classes.

# HTML-related output production.

html_header = """<?xml version="1.0" encoding="iso-8859-15"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>Module</title>
  <style type="text/css">
    html {
      background-color: black; color: white;
    }

    body {
      padding-bottom: 4em;
      font-size: 14pt; font-family: monospace;
      background-color: black; color: white;
    }

    .nowrap { white-space: nowrap; }

    .class { margin-top: 1em; margin-bottom: 1em; }
    .function { margin-top: 1em; margin-bottom: 1em; }
    .body { padding-left: 2em; }
    .keyword { color: yellow; }
    .comment { color: blue; }
    .class-name { color: cyan; }
    .function-name { color: cyan; }
    .str { color: #FF00FF; }
    .doc { color: #FF00FF; margin-top: 1em; margin-bottom: 1em; }
    .invocation a { color: white; text-decoration: none; }

    .summary-class {
      background-color: #004400;
    }

    .summary-instance {
      background-color: #0000FF;
    }

    .summary-attr {
      background-color: #007700;
    }

  </style>
</head>
<body>
"""

html_footer = """</body>
</html>
"""

# Utility classes.

class Writer:

    "A utility class providing useful HTML output methods."

    # Methods which return strings.

    def _text(self, text):
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    def _attr(self, attr):
        return self._text(attr).replace("'", "&apos;").replace('"', "&quot;")

    def _url(self, url):
        return self._attr(url).replace("#", "%23").replace("-", "%2d")

    # Methods which write to the stream.

    def _comment(self, comment):
        self.stream.write("<span class='comment'># %s</span>\n" % comment)

    def _keyword(self, kw):
        self.stream.write("<span class='keyword'>%s</span> " % kw)

    def _doc(self, node):
        if node.doc is not None:
            self.stream.write("<pre class='doc'>\n")
            self.stream.write('"""')
            output = textwrap.dedent(node.doc.replace('"""', '\\"\\"\\"'))
            self.stream.write(self._text(output))
            self.stream.write('"""')
            self.stream.write("</pre>\n")

    def _name(self, name, classes=None):
        self.stream.write("<span class='%s'>%s</span>" % (classes or "name", name))

    def _name_link(self, module_name, full_name, name, classes=None):
        self.stream.write("<a class='%s' href='%s%sxhtml#%s'>%s</a>" % (
            classes or "name", module_name, os.path.extsep,
            self._attr(full_name), self._text(name)))

    def _summary_link(self, module_name, full_name, name, classes=None):
        self._name_link("%s-summary" % module_name, full_name, name, classes)

    def _object_name_def(self, module, obj, classes=None):

        """
        Link to the summary for 'module' using 'obj'. The optional 'classes'
        can be used to customise the CSS classes employed.
        """

        self._summary_link(module.full_name(), obj.full_name(), obj.name, classes)

# Summary classes.

class Summary(Writer):

    "Summarise classes and attributes in modules."

    def __init__(self, module):
        self.module = module

    def to_stream(self, stream):

        "Write the summary to the given 'stream'."

        self.stream = stream
        self._init_details()
        self.stream.write(html_header)
        self._write_classes(self.module)
        self.stream.write(html_footer)

    def _write_classes(self, module):
        self.stream.write("<table cellspacing='5' cellpadding='5'>\n")

        all_classes = {}

        for obj in self.module.all_objects:
            if isinstance(obj, Class):
                all_classes[obj.name] = obj

        all_class_names = all_classes.keys()
        all_class_names.sort()

        for name in all_class_names:
            self._write_class(all_classes[name])

        self.stream.write("</table>\n")

    def _write_class(self, obj):

        # Write the class...

        self.stream.write("<tbody class='class'>\n")
        self.stream.write("<tr>\n")
        self.stream.write("<th class='summary-class' id='%s'>\n" % obj.full_name())
        self._keyword("class")
        self.stream.write(obj.name)
        self.stream.write("</th>\n")

        # ...and all known attribute names.

        obj_attributes = obj.all_attribute_names()

        for name in self.attribute_names:
            if name in obj_attributes:
                self.stream.write("<th class='summary-attr'>%s</th>\n" % self._text(name))
            else:
                self.stream.write("<th></th>\n")
        self.stream.write("</tr>\n")

        self.stream.write("</tbody>\n")

    def _init_details(self):
        names = set()

        # Visit all classes.

        for obj in self.module.all_objects:
            if isinstance(obj, Class):
                names.update(obj.all_attribute_names())

        self.attribute_names = list(names)
        self.attribute_names.sort()

# Source code classes.

class AnnotatedSource(ASTVisitor, Writer):

    "A module source code browser."

    def __init__(self, module):
        ASTVisitor.__init__(self)
        self.visitor = self
        self.module = module

    def to_stream(self, stream):

        "Write the annotated code to the given 'stream'."

        self.stream = stream
        self.stream.write(html_header)
        self.dispatch(self.module.module)
        self.stream.write(html_footer)

    def visitModule(self, node):
        self.default(node)

    # Statements.

    def visitAssert(self, node):
        self.stream.write("<div class='assert nowrap'>\n")
        self._keyword("assert")
        self.dispatch(node.test)
        if node.fail:
            self.stream.write(", ")
            self.dispatch(node.fail)
        self.stream.write("</div>\n")

    def visitAssign(self, node):
        self.stream.write("<div class='assign nowrap'>\n")
        for lvalue in node.nodes:
            self.dispatch(lvalue)
            self.stream.write("= ")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitAugAssign(self, node):
        self.stream.write("<div class='augassign nowrap'>\n")
        self.dispatch(node.node)
        self.stream.write("%s " % node.op)
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitBreak(self, node):
        self.stream.write("<div class='break nowrap'>\n")
        self._keyword("break")
        self.stream.write("</div>\n")

    def visitClass(self, node):

        # Use inspected details where possible.

        if hasattr(node, "_def"):
            cls = node._def
            bases = cls.bases
            self.stream.write("<div class='class nowrap' id='%s'>\n" % cls.full_name())
        else:
            print "Warning: class %s not recognised!" % node.name
            return

        # Write the declaration line.

        self.stream.write("<div>\n")
        self._keyword("class")
        self._object_name_def(self.module, cls, "class-name")

        # Suppress the "object" class appearing alone.

        if bases and not (len(bases) == 1 and bases[0].name == "object"):
            self.stream.write("(")
            first = 1
            for base in bases:
                if not first:
                    self.stream.write(", ")

                self._object_name_def(base.module, base)

                first = 0
            self.stream.write(")")

        self.stream.write(":\n")
        self.stream.write("</div>\n")

        # Write the docstring and class body.

        self.stream.write("<div class='body nowrap'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitContinue(self, node):
        self.stream.write("<div class='continue nowrap'>\n")
        self._keyword("continue")
        self.stream.write("</div>\n")

    def visitDiscard(self, node):
        self.stream.write("<div class='discard nowrap'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitFor(self, node):
        self.stream.write("<div class='if nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("for")
        self.dispatch(node.assign)
        self._keyword("in")
        self.dispatch(node.list)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitFrom(self, node):
        self.stream.write("<div class='from nowrap'>\n")
        self._keyword("from")
        self._name(node.modname)
        self._keyword("import")
        first = 1
        for (name, alias), _name in map(None, node.names, node._names):
            if not first:
                self.stream.write(", ")
            if alias:
                self.stream.write(name + " ")
                self._keyword("as")
            self._name(alias or name)
            first = 0
        self.stream.write("</div>\n")

    def visitFunction(self, node):
        if hasattr(node, "_def"):
            fn = node._def
            self.stream.write("<div class='function nowrap' id='%s'>\n" % fn.full_name())
        else:
            print "Warning: function %s not recognised!" % node.name
            return

        # Write the declaration line.

        self.stream.write("<div>\n")
        self._keyword("def")
        self._object_name_def(self.module, fn, "function-name")

        self.stream.write("(")
        self._parameters(fn)
        self.stream.write(")")
        self.stream.write(":\n")
        self.stream.write("</div>\n")

        self.stream.write("<div class='body nowrap'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitGlobal(self, node):
        self.stream.write("<div class='global nowrap'>\n")
        self._keyword("global")
        first = 1
        for name in node.names:
            if not first:
                self.stream.write(", ")
            self.stream.write(name)
            first = 0
        self.stream.write("</div>\n")

    def visitIf(self, node):
        self.stream.write("<div class='if nowrap'>\n")
        first = 1
        for compare, stmt in node.tests:
            self.stream.write("<div>\n")
            if first:
                self._keyword("if")
            else:
                self._keyword("elif")
            self.dispatch(compare)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(stmt)
            self.stream.write("</div>\n")
            first = 0
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    # Output preparation methods.

    def _sequence(self, node):
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(", ")
            self.dispatch(n)
            first = 0

    def _mapping(self, node):
        first = 1
        for k, v in node.items:
            if not first:
                self.stream.write(", ")
            self.dispatch(k)
            self.stream.write(" : ")
            self.dispatch(v)
            first = 0

    def _parameters(self, fn):
        nparams = len(fn.positional_names)
        ndefaults = len(fn.defaults)
        first_with_default = nparams - ndefaults

        first = 1
        for n, param in enumerate(fn.positional_names):
            if not first:
                self.stream.write(", ")
            self._name(param)
            n_default = n - first_with_default
            if n_default >= 0:
                self._default(fn.defaults[n_default])
            first = 0

        if fn.has_star:
            if not first:
                self.stream.write(", *")
            self._name(fn.star_name)

        if fn.has_dstar:
            if not first:
                self.stream.write(", **")
            self._name(fn.dstar_name)

    def _default(self, default):
        self.stream.write("=")
        self.dispatch(default)

# Convenience functions.

def summarise(module, filename):
    stream = open(filename, "wb")
    try:
        summary = Summary(module)
        summary.to_stream(stream)
    finally:
        stream.close()

def annotate(module, filename):
    stream = open(filename, "wb")
    try:
        source = AnnotatedSource(module)
        source.to_stream(stream)
    finally:
        stream.close()

# vim: tabstop=4 expandtab shiftwidth=4
